/* umax.c: scanner driver for UMAX scanners.
 * Copyright (C) 1995, 1996, 1997 Michael K. Johnson
 * Tested on my Vista-S6 (Envisions 6600S); should work on all
 * UMAX SCSI-2 scanners.  If it doesn't, please tell me and we'll
 * work it out, I hope...
 * 
 * I can be reached via email at johnsonm@redhat.com
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 * 
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <signal.h>
#include <stdlib.h>

/* Get into kernel source tree... */
/* first need to create a dummy version of this structure that scsi.h wants.
 * Won't hurt us, just lets the compiler continue... */
struct request {
  char dummy;
};
#include <linux/../scsi/scsi.h>
#include <linux/../scsi/sg.h>


#include "scsicmd.h" /* must come first; umax.h requires it */
#include "umax.h"
#include "debug.h"



#define min(a,b) (((a)<(b))?(a):(b))

typedef struct umax_scanner {
  int (*grab_scanner)(struct umax_scanner *us);
  int (*do_scan)(struct umax_scanner *us);
  int (*give_scanner)(struct umax_scanner *us);
  unsigned char buffer[BUFSIZE];
  #define bufcom(buf) (buf + SCSI_OFF)
  unsigned int bufsize;
  unsigned char *arena; /* used to mmap files into for writing */
  int arena_offset;     /* used to indicate arena offset for graphic data */
  char *devicename;
  int verbose;
  int colormode; /* LINEART, GREYSCALE, or RGB */
  #define LINEART 1
  #define GREYSCALE 2
  #define RGB 3
  int adf; /* 1 if ADF turned on */
  int row_len;
  int width_in_pixels;
  int height_in_pixels;
  int phys_width; /* 1200pt/inch */
  int phys_height; /* 1200pt/inch */
  unsigned int max_value;
  unsigned int x_resolution;
  unsigned int y_resolution;
  unsigned int upper_left_x; /* 1200pt/inch */
  unsigned int upper_left_y; /* 1200pt/inch */
  int brightness;
  int threshold;
  int ofd; /* output file descriptor, stdout by default */
} US;


/* wait_scanner should spin until TEST_INIT_READY returns GOOD
   returns 0 on success,
   returns -1 on error.  */

int wait_scanner(US *us) {
  int ret;
  int cnt = 0;

  do {
    scsicpy(us->buffer, test_unit_ready.cmd, test_unit_ready.size);
    ret = do_scsi_cmd(test_unit_ready.size, us->buffer, 0, NULL, -1);
    cnt++;
    if (cnt > 100) {
      fprintf(stderr, "testing unit; not ready\n");
      return -1;
    }
    if (ret == BUSY) {
      usleep(100000); /* .1 seconds */
    }
  } while (ret != GOOD);
  return 0;
}

#define WAIT_SCANNER ret = wait_scanner(us); if (ret) return ret

/* umax_grab_scanner should go through the following command sequence:
 * TEST UNIT READY
 *     CHECK CONDITION  \
 * REQUEST SENSE         > These should be handled automagically by
 *     UNIT ATTENTION   /  the kernel if they happen (powerup/reset)
 * TEST UNIT READY
 *     GOOD
 * RESERVE UNIT
 *     GOOD
 * 
 * It is then responsible for installing appropriate signal handlers
 * to call emergency_give_scanner() if user aborts.
 */
static US *emergency_us;
int emergency_give_scanner(int i) {

  if (emergency_us->verbose) {
    fprintf(stderr, "emergency: trying to release scanner ...");
    fflush(stderr);
  }
  fullcmd(emergency_us->buffer, object_position, 0, NULL, GOOD);
  wait_scanner(emergency_us);
  fullcmd(emergency_us->buffer, release_unit, 0, NULL, GOOD);
  if (emergency_us->verbose) {
    fprintf(stderr, " scanner released\n");
    fflush(stderr);
  }
  exit(1);
}

int umax_grab_scanner(US *us) {
  int ret;
  int cnt = 0;

  do {
    scsicpy(us->buffer, test_unit_ready.cmd, test_unit_ready.size);
    ret = do_scsi_cmd(test_unit_ready.size, us->buffer, 0, NULL, -1);
    cnt++;
    if (cnt > 100) {
      fprintf(stderr, "testing unit; not ready\n");
      return -1;
    }
    if (ret == BUSY) {
      /* RS_sense_key_warmup is not implemented.
       * (we can simply assume that if it is busy we try back again
       * in a little bit...)
       * The test would be:
       * (RS_sense_key(((struct sg_header)(us->buffer))->sense_buffer) & RS_sense_key_warmup)
       */
      usleep(100000); /* .1 seconds */
    }
  } while (ret != GOOD);
  fullcmd(us->buffer, reserve_unit, 0, NULL, GOOD);
  if (us->verbose) {
    fprintf(stderr, "unit tested ready after %d tries; unit reserved\n", cnt);
  }
  /* And if we abort? */
  emergency_us = us;
  signal(SIGINT, (__sighandler_t) emergency_give_scanner);
  signal(SIGTERM, (__sighandler_t) emergency_give_scanner);
  signal(SIGABRT, (__sighandler_t) emergency_give_scanner);
  signal(SIGILL, (__sighandler_t) emergency_give_scanner);
  signal(SIGSEGV, (__sighandler_t) emergency_give_scanner);
  return 0;
}


/* umax_reposition_scanner is used on three-pass scanners where
 * we want to reset the carriage without releasing the bus
 */
int umax_reposition_scanner(US *us) {
  int ret;

  if (us->verbose) {
    fprintf(stderr, "trying to reposition scanner ...");
    fflush(stderr);
  }
  fullcmd(us->buffer, object_position, 0, NULL, GOOD);
  WAIT_SCANNER;
  if (us->verbose) {
    fprintf(stderr, " scanner repositioned\n");
    fflush(stderr);
  }
  return 0;
}

/* umax_give_scanner should go through the following sequence:
 * OBJECT POSITION
 *     GOOD
 * RELEASE UNIT
 *     GOOD
 */

int umax_give_scanner(US *us) {
  int ret;

  if (us->verbose) {
    fprintf(stderr, "trying to release scanner ...");
    fflush(stderr);
  }
  fullcmd(us->buffer, object_position, 0, NULL, GOOD);
  WAIT_SCANNER;
  fullcmd(us->buffer, release_unit, 0, NULL, GOOD);
  if (us->verbose) {
    fprintf(stderr, " scanner released\n");
    fflush(stderr);
  }
  return 0;
}




/* set_window_param sets all the window parameters.  This
 * means building a fairly complicated SCSI command before
 * sending it...
 */
int vista_s6_set_window_param(US *us) {
  int num_dblocks = 1; /* number of window descriptor blocks, usually 1 or 3 */
  int i;

#ifndef LONG_WDB
  /* for now, hard-coded to length for vista-s6 */
  set_WDB_length_vista_s6;
#endif
  if (us->colormode == RGB) num_dblocks = 3;
  scsicpy(us->buffer, set_window.cmd, set_window.size);
  scsicpy(WPDB_OFF(bufcom(us->buffer)),
	  window_parameter_data_block.cmd, window_parameter_data_block.size);
  set_WPDB_wdbnum(WPDB_OFF(bufcom(us->buffer)), num_dblocks);
  for (i = 1; i <= num_dblocks; i++) {
    scsicpy(WDB_OFF(us->buffer, i),
	    window_descriptor_block.cmd, window_descriptor_block.size);
    /* If there is only one WDB, then the WID is 0.
     * Otherwise, the WIDs are 1-3.  Yuck. */
    set_WD_wid(WDB_OFF(bufcom(us->buffer),i), (num_dblocks > 1)?i:0);
    set_WD_Xres(WDB_OFF(bufcom(us->buffer),i), us->x_resolution);
    set_WD_Yres(WDB_OFF(bufcom(us->buffer),i), us->y_resolution);
    set_WD_ULX(WDB_OFF(bufcom(us->buffer),i), us->upper_left_x);
    set_WD_ULY(WDB_OFF(bufcom(us->buffer),i), us->upper_left_y);
    set_WD_width(WDB_OFF(bufcom(us->buffer),i), us->phys_width);
    set_WD_length(WDB_OFF(bufcom(us->buffer),i), us->phys_height);
    set_WD_brightness(WDB_OFF(bufcom(us->buffer),i), us->brightness);
    set_WD_threshold(WDB_OFF(bufcom(us->buffer),i), us->threshold);
#ifdef LONG_WDB
    set_WD_pixel_count(WDB_OFF(bufcom(us->buffer), i), us->width_in_pixels);
    set_WD_line_count(WDB_OFF(bufcom(us->buffer), i), us->height_in_pixels);
#endif
    switch(us->colormode) {
     case LINEART:
      set_WD_composition(WDB_OFF(bufcom(us->buffer), i), WD_comp_lineart);
      set_WD_bitsperpixel(WDB_OFF(bufcom(us->buffer), i), WD_bits_1);
      set_WD_select_color(WDB_OFF(bufcom(us->buffer), i), WD_color_grey);
#ifdef LONG_WDB
      set_WD_calibration(WDB_OFF(bufcom(us->buffer), i), WD_calibration_lineart);
#endif
      break;
     case GREYSCALE:
      set_WD_composition(WDB_OFF(bufcom(us->buffer), i), WD_comp_grey);
      set_WD_bitsperpixel(WDB_OFF(bufcom(us->buffer), i), WD_bits_8);
      set_WD_select_color(WDB_OFF(bufcom(us->buffer), i), WD_color_grey);
#ifdef LONG_WDB
      set_WD_calibration(WDB_OFF(bufcom(us->buffer), i), WD_calibration_grey);
#endif
      break;
     case RGB:
      set_WD_composition(WDB_OFF(bufcom(us->buffer), i), WD_comp_rgb_full);
      set_WD_bitsperpixel(WDB_OFF(bufcom(us->buffer), i), WD_bits_8);
      switch (i) {
       /* first red, then green, then blue */
       case 1:
	set_WD_select_color(WDB_OFF(bufcom(us->buffer), i), WD_color_red);
	break;
       case 2:
	set_WD_select_color(WDB_OFF(bufcom(us->buffer), i), WD_color_green);
	break;
       case 3:
	set_WD_select_color(WDB_OFF(bufcom(us->buffer), i), WD_color_blue);
	break;
      }
#ifdef LONG_WDB
      set_WD_calibration(WDB_OFF(bufcom(us->buffer), i), WD_calibration_rgb);
#endif
      break;
    }
    if (us->verbose) {
      fprintf(stderr, "window descriptor block %d created\n", i);
    }
  }

  /* manipulate the command and window descriptor block(s) in the
   * buffer here */

  set_SW_xferlen(bufcom(us->buffer), (window_parameter_data_block.size +
			  (window_descriptor_block.size * num_dblocks)));
  docmd(set_window.size + window_parameter_data_block.size +
	(window_descriptor_block.size * num_dblocks), us->buffer, 0, NULL, GOOD);
  if (us->verbose) {
    fprintf(stderr, "window(s) set\n");
  }
  return 0;
}

int vista_s6_start_scan(US *us) {
  char typebuf[256];
  int widlistlen = 0; /* length of window id list */
#if 0
  int i;
#endif

  scsicpy(us->buffer, scan.cmd, scan.size);
#if 0
  /* The Envisions driver for the Vista S6 ALWAYS sets widlistlen
   * to 1 and makes the list element be a 0.  This works.
   * When I try to do that, the scanner aborts.  I don't know why.
   * At least for color and greyscale, setting widlistlen to 0
   * and having no list at all works.  <sigh>.  If anyone figures
   * this one out, please let me know.  If anyone wants to attack
   * the problem, I can send you traces.
   */
  switch(us->colormode) {
   case LINEART:
   case GREYSCALE:
    set_SC_wid(bufcom(us->buffer), 1, 0);
    widlistlen = 1;
    break;
   case RGB:
    for (i = 1; i <=3; i++) {
      set_SC_wid(bufcom(us->buffer), i, i);
    }
    widlistlen = 3;
    break;
  }
  /* manipulate the command in the buffer here to set
   * quality, use of ADF, and/or the preview bit */
  set_SC_xfer_length(bufcom(us->buffer), widlistlen);
#endif
  set_SC_adf(bufcom(us->buffer), us->adf);
  docmd(scan.size + widlistlen, us->buffer, 0, NULL, GOOD);
  /* Now put out the appropriate header.
   * With the appropriate header, it looks like raw scanner data is
   * the appropriate format below, so no munging needs to take place.
   * If 30-bit format is added, it will have to be output in *ascii*
   * mode to use pnm utilities.  Yuck!
   */
  switch(us->colormode) {
   case LINEART:
    sprintf(typebuf, "P4\n# pbm created by umax\n%u %u\n",
	    us->width_in_pixels, us->height_in_pixels);
    break;
   case GREYSCALE:
    sprintf(typebuf, "P5\n# pgm created by umax\n%u %u\n%u\n",
	    us->width_in_pixels, us->height_in_pixels, us->max_value);
    break;
   case RGB:
    sprintf(typebuf, "P6\n# ppm created by umax\n%u %u\n%u\n",
	    us->width_in_pixels, us->height_in_pixels, us->max_value);
    break;
  }
  write(us->ofd, typebuf, strlen(typebuf));
  return 0;
}


/* For scanners that need neatly-sized reads, use code like this:
 * row_bufsize = trim_bufsize(us->bufsize, us->row_len);
 */
int trim_bufsize(bufsize, row_len) {
  return (bufsize < row_len) ? bufsize
                             : bufsize - (bufsize % row_len);
}


int vista_s6_read_data(US *us) {
  unsigned int data_left = 0;
  unsigned int data_to_read;
  unsigned int row_bufsize;
  int ret;

  if (us->verbose) {
    fprintf(stderr, "Starting to READ data\n");
  }
  switch(us->colormode) {
   case LINEART:
    data_left = (us->width_in_pixels * us->height_in_pixels / 8);
    break;
   case GREYSCALE:
    /* one byte per pixel */
    data_left = us->width_in_pixels * us->height_in_pixels;
    break;
   case RGB:
    /* three bytes per pixel.  Note that high-end scanners
     * might have 6, if they are 30 bits deep total, since
     * they spread each pixel over two bytes.  No weird packing!
     * If you have one of these scanners, please work with me
     * to extend this program to deal with your extra bits.
     */
    data_left = us->width_in_pixels * us->height_in_pixels * 3;
    break;
  }

  row_bufsize = us->bufsize;
  if (us->verbose) {
    fprintf(stderr, "reading %u bytes in rows of %u\n",
	    data_left, row_bufsize);
  }

  do {
    /* if we want to, we can check to see if data_left agrees
     * with amount returned by GET DATA BUFFER STATUS, but why
     * bother? */

    data_to_read = (data_left < row_bufsize) ? data_left : row_bufsize;

    WAIT_SCANNER;
    scsicpy(us->buffer, sread.cmd, sread.size);
    set_R_xfer_length(bufcom(us->buffer), data_to_read);
    ret = do_scsi_cmd(sread.size, us->buffer, data_to_read, us->buffer, GOOD);
    if (ret < 0) {
      if (ret == -4) {
	/* changed to smaller buffer size; try again */
        us->bufsize = -1; /* don't set bufsize back to original size! */
	scsi_init(0, &(us->bufsize));
	row_bufsize = us->bufsize;
	if (us->verbose) {
	  fprintf(stderr, "Retrying read: reading %u bytes in rows of %u\n",
		  data_left, row_bufsize);
	}
	continue;
      } else {
	/* some other non-recoverable error */
	return -1;
      }
    }
    /* output this chunk of data, which just happens to be
     * in pnm data format.  Yay! */
    write(us->ofd, bufcom(us->buffer), data_to_read);

    data_left -= data_to_read;
    if (us->verbose) {
      fprintf(stderr, "Buffer of %d bytes read; %d bytes to go\n",
              data_to_read, data_left);
    }
  } while (data_left);
  return 0;
}


int vista_s12_set_window_param(US *us) {
  int num_dblocks = 1; /* number of window descriptor blocks, usually 1 or 3 */
  int i;

#ifndef LONG_WDB
  /* for now, hard-coded to length for vista-s12 */
  set_WDB_length_vista_s12;
#endif
  if (us->colormode == RGB) num_dblocks = 3;
  scsicpy(us->buffer, set_window.cmd, set_window.size);
  scsicpy(WPDB_OFF(bufcom(us->buffer)),
	  window_parameter_data_block.cmd, window_parameter_data_block.size);
  set_WPDB_wdbnum(WPDB_OFF(bufcom(us->buffer)), num_dblocks);
  for (i = 1; i <= num_dblocks; i++) {
    scsicpy(WDB_OFF(us->buffer, i),
	    window_descriptor_block.cmd, window_descriptor_block.size);
    /* If there is only one WDB, then the WID is 0.
     * Otherwise, the WIDs are 1-3.  Yuck. */
    set_WD_wid(WDB_OFF(bufcom(us->buffer),i), (num_dblocks > 1)?i:0);
    set_WD_Xres(WDB_OFF(bufcom(us->buffer),i), us->x_resolution);
    set_WD_Yres(WDB_OFF(bufcom(us->buffer),i), us->y_resolution);
    set_WD_ULX(WDB_OFF(bufcom(us->buffer),i), us->upper_left_x);
    set_WD_ULY(WDB_OFF(bufcom(us->buffer),i), us->upper_left_y);
    set_WD_width(WDB_OFF(bufcom(us->buffer),i), us->phys_width);
    set_WD_length(WDB_OFF(bufcom(us->buffer),i), us->phys_height);
    set_WD_brightness(WDB_OFF(bufcom(us->buffer),i), us->brightness);
    set_WD_threshold(WDB_OFF(bufcom(us->buffer),i), us->threshold);
#ifdef LONG_WDB
    set_WD_pixel_count(WDB_OFF(bufcom(us->buffer), i), us->width_in_pixels);
    set_WD_line_count(WDB_OFF(bufcom(us->buffer), i), us->height_in_pixels);
#endif
    switch(us->colormode) {
     case LINEART:
      set_WD_composition(WDB_OFF(bufcom(us->buffer), i), WD_comp_lineart);
      set_WD_bitsperpixel(WDB_OFF(bufcom(us->buffer), i), WD_bits_1);
      set_WD_select_color(WDB_OFF(bufcom(us->buffer), i), WD_color_grey);
#ifdef LONG_WDB
      set_WD_calibration(WDB_OFF(bufcom(us->buffer), i), WD_calibration_lineart);
#endif
      break;
     case GREYSCALE:
      set_WD_composition(WDB_OFF(bufcom(us->buffer), i), WD_comp_grey);
      set_WD_bitsperpixel(WDB_OFF(bufcom(us->buffer), i), WD_bits_8);
      set_WD_select_color(WDB_OFF(bufcom(us->buffer), i), WD_color_grey);
#ifdef LONG_WDB
      set_WD_calibration(WDB_OFF(bufcom(us->buffer), i), WD_calibration_grey);
#endif
      break;
     case RGB:
      set_WD_composition(WDB_OFF(bufcom(us->buffer), i), WD_comp_rgb_full);
      set_WD_bitsperpixel(WDB_OFF(bufcom(us->buffer), i), WD_bits_8);
      switch (i) {
       /* first red, then green, then blue */
       case 1:
	set_WD_select_color(WDB_OFF(bufcom(us->buffer), i), WD_color_red);
	break;
       case 2:
	set_WD_select_color(WDB_OFF(bufcom(us->buffer), i), WD_color_green);
	break;
       case 3:
	set_WD_select_color(WDB_OFF(bufcom(us->buffer), i), WD_color_blue);
	break;
      }
#ifdef LONG_WDB
      set_WD_calibration(WDB_OFF(bufcom(us->buffer), i), WD_calibration_rgb);
#endif
      break;
    }
    if (us->verbose) {
      fprintf(stderr, "window descriptor block %d created\n", i);
    }
  }

  /* manipulate the command and window descriptor block(s) in the
   * buffer here */

  set_SW_xferlen(bufcom(us->buffer), (window_parameter_data_block.size +
			  (window_descriptor_block.size * num_dblocks)));
  docmd(set_window.size + window_parameter_data_block.size +
	(window_descriptor_block.size * num_dblocks), us->buffer, 0, NULL, GOOD);
  if (us->verbose) {
    fprintf(stderr, "window(s) set\n");
  }
  return 0;
}

int vista_s12_start_scan(US *us) {
  char typebuf[256];
  int widlistlen = 0; /* length of window id list */
#if 0
  int i;
#endif

  scsicpy(us->buffer, scan.cmd, scan.size);
#if 0
  /* The Envisions driver for the Vista S6 ALWAYS sets widlistlen
   * to 1 and makes the list element be a 0.  This works.
   * When I try to do that, the scanner aborts.  I don't know why.
   * At least for color and greyscale, setting widlistlen to 0
   * and having no list at all works.  <sigh>.  If anyone figures
   * this one out, please let me know.  If anyone wants to attack
   * the problem, I can send you traces.
   */
  switch(us->colormode) {
   case LINEART:
   case GREYSCALE:
    set_SC_wid(bufcom(us->buffer), 1, 0);
    widlistlen = 1;
    break;
   case RGB:
    for (i = 1; i <=3; i++) {
      set_SC_wid(bufcom(us->buffer), i, i);
    }
    widlistlen = 3;
    break;
  }
  /* manipulate the command in the buffer here to set
   * quality, use of ADF, and/or the preview bit */
  set_SC_xfer_length(bufcom(us->buffer), widlistlen);
#endif
  set_SC_adf(bufcom(us->buffer), us->adf);
  docmd(scan.size + widlistlen, us->buffer, 0, NULL, GOOD);
  /* Now put out the appropriate header.
   * With the appropriate header, it looks like raw scanner data is
   * the appropriate format below, so no munging needs to take place.
   * If 30-bit format is added, it will have to be output in *ascii*
   * mode to use pnm utilities.  Yuck!
   */
  switch(us->colormode) {
   case LINEART:
    sprintf(typebuf, "P4\n# pbm created by umax\n%u %u\n",
	    us->width_in_pixels, us->height_in_pixels);
    break;
   case GREYSCALE:
    sprintf(typebuf, "P5\n# pgm created by umax\n%u %u\n%u\n",
	    us->width_in_pixels, us->height_in_pixels, us->max_value);
    break;
   case RGB:
    sprintf(typebuf, "P6\n# ppm created by umax\n%u %u\n%u\n",
	    us->width_in_pixels, us->height_in_pixels, us->max_value);
    break;
  }
  write(us->ofd, typebuf, strlen(typebuf));
  return 0;
}

int vista_s12_read_data(US *us) {
  unsigned int data_left = 0;
  unsigned int data_to_read;
  unsigned int row_bufsize;
  int ret;

  if (us->verbose) {
    fprintf(stderr, "Starting to READ data\n");
  }
  switch(us->colormode) {
   case LINEART:
    data_left = (us->width_in_pixels * us->height_in_pixels / 8);
    break;
   case GREYSCALE:
    /* one byte per pixel */
    data_left = us->width_in_pixels * us->height_in_pixels;
    break;
   case RGB:
    /* three bytes per pixel.  Note that high-end scanners
     * might have 6, if they are 30 bits deep total, since
     * they spread each pixel over two bytes.  No weird packing!
     * If you have one of these scanners, please work with me
     * to extend this program to deal with your extra bits.
     */
    data_left = us->width_in_pixels * us->height_in_pixels * 3;
    break;
  }

  row_bufsize = us->bufsize;
  if (us->verbose) {
    fprintf(stderr, "reading %u bytes in rows of %u\n",
	    data_left, row_bufsize);
  }

  do {
    /* if we want to, we can check to see if data_left agrees
     * with amount returned by GET DATA BUFFER STATUS, but why
     * bother? */

    data_to_read = (data_left < row_bufsize) ? data_left : row_bufsize;

    WAIT_SCANNER;
    scsicpy(us->buffer, sread.cmd, sread.size);
    set_R_xfer_length(bufcom(us->buffer), data_to_read);
    ret = do_scsi_cmd(sread.size, us->buffer, data_to_read, us->buffer, GOOD);
    if (ret < 0) {
      if (ret == -4) {
	/* changed to smaller buffer size; try again */
        us->bufsize = -1; /* don't set bufsize back to original size! */
	scsi_init(0, &(us->bufsize));
	row_bufsize = us->bufsize;
	if (us->verbose) {
	  fprintf(stderr, "Retrying read: reading %u bytes in rows of %u\n",
		  data_left, row_bufsize);
	}
	continue;
      } else {
	/* some other non-recoverable error */
	return -1;
      }
    }
    /* output this chunk of data, which just happens to be
     * in pnm data format.  Yay! */
    write(us->ofd, bufcom(us->buffer), data_to_read);

    data_left -= data_to_read;
    if (us->verbose) {
      fprintf(stderr, "Buffer of %d bytes read; %d bytes to go\n",
              data_to_read, data_left);
    }
  } while (data_left);
  return 0;
}


int t630_write_header(US *us, int color) {
  char typebuf[256];
  int buflen = 0;

  switch(us->colormode) {
   case LINEART:
    buflen = sprintf(typebuf, "P4\n# pbm created by umax\n%u %u\n",
	             us->width_in_pixels, us->height_in_pixels);
    break;
   case GREYSCALE:
    buflen = sprintf(typebuf, "P5\n# pgm created by umax\n%u %u\n%u\n",
	             us->width_in_pixels, us->height_in_pixels, us->max_value);
    break;
   case RGB:
    if (us->arena) {
      if (color != 1) /* this only needs to happen once here */
	return 0;
      buflen = sprintf(typebuf, "P6\n# ppm created by umax\n%u %u\n%u\n",
	               us->width_in_pixels, us->height_in_pixels, us->max_value);
      break;
    } else { /* no memory mapping; need to write three images out if color */
      buflen = sprintf(typebuf, "P5\n# pgm as ppm part %d created by umax; convert to ppm with threetoimage\n%u %u\n%u\n",
	               color, us->width_in_pixels, us->height_in_pixels, us->max_value);
      break;
    }
  }
  if (us->arena) {
    strncpy(us->arena, typebuf, buflen);
    us->arena_offset = buflen;
  } else {
    write(us->ofd, typebuf, buflen);
  }
  return 0;
}

int t630_write_data(US *us, int color, char *data, int len, int where) {
  int i;

  if (us->arena) {
    switch(us->colormode) {
      case LINEART:
      case GREYSCALE:
       memcpy(us->arena + us->arena_offset + where, data, len);
       break;
      case RGB:
       for (i=0; i<len; i++) {
         /* stripe the data into the output file */
         /* The mapped file plus the offset for the header... */
         *((us->arena + us->arena_offset)
           + (where*3) + (color-1) + (i*3)) = data[i];
         /*   plus the current place, plus the color offset,
          *   plus the data element offset times the stripe size,
          *                                 is set to the data element */
       }
       break;
    }
  } else { /* no mmapping, so just write. if RGB, three images are going out */
    write(us->ofd, data, len);
  }
  return 0;
}


/* set_window_param sets all the window parameters.  This
 * means building a fairly complicated SCSI command before
 * sending it...
 */
int t630_set_window_param(US *us, int color) {

  set_WDB_length_t630;
  memset(us->buffer, '\0', BUFSIZE);
  scsicpy(us->buffer, set_window.cmd, set_window.size);
  scsicpy(WPDB_OFF(bufcom(us->buffer)),
	  window_parameter_data_block.cmd, window_parameter_data_block.size);
  set_WPDB_wdbnum(WPDB_OFF(bufcom(us->buffer)), 1);
  scsicpy(WDB_OFF(us->buffer, 1),
	  window_descriptor_block.cmd, window_descriptor_block.size);
  set_WD_wid(WDB_OFF(bufcom(us->buffer),1), 0);
  set_WD_Xres(WDB_OFF(bufcom(us->buffer),1), us->x_resolution);
  set_WD_Yres(WDB_OFF(bufcom(us->buffer),1), us->y_resolution);
  set_WD_ULX(WDB_OFF(bufcom(us->buffer),1), us->upper_left_x);
  set_WD_ULY(WDB_OFF(bufcom(us->buffer),1), us->upper_left_y);
  set_WD_width(WDB_OFF(bufcom(us->buffer),1), us->phys_width);
  set_WD_length(WDB_OFF(bufcom(us->buffer),1), us->phys_height);
  set_WD_brightness(WDB_OFF(bufcom(us->buffer),1), us->brightness);
  set_WD_threshold(WDB_OFF(bufcom(us->buffer),1), us->threshold);
  switch(us->colormode) {
   case LINEART:
    set_WD_composition(WDB_OFF(bufcom(us->buffer), 1), WD_comp_lineart);
    set_WD_bitsperpixel(WDB_OFF(bufcom(us->buffer), 1), WD_bits_1);
    set_WD_select_color(WDB_OFF(bufcom(us->buffer), 1), WD_color_grey);
    break;
   case GREYSCALE:
    set_WD_composition(WDB_OFF(bufcom(us->buffer), 1), WD_comp_grey);
    set_WD_bitsperpixel(WDB_OFF(bufcom(us->buffer), 1), WD_bits_8);
    set_WD_select_color(WDB_OFF(bufcom(us->buffer), 1), WD_color_grey);
    break;
   case RGB:
    set_WD_composition(WDB_OFF(bufcom(us->buffer), 1), WD_comp_rgb_full);
    set_WD_bitsperpixel(WDB_OFF(bufcom(us->buffer), 1), WD_bits_8);
    switch (color) {
     case 1:
      set_WD_select_color(WDB_OFF(bufcom(us->buffer), 1), WD_color_red);
      break;
     case 2:
      set_WD_select_color(WDB_OFF(bufcom(us->buffer), 1), WD_color_green);
      break;
     case 3:
      set_WD_select_color(WDB_OFF(bufcom(us->buffer), 1), WD_color_blue);
      break;
    }
    break;
  }
  if (us->verbose) {
    fprintf(stderr, "window descriptor block %d created\n", color);
  }

  /* manipulate the command and window descriptor block(s) in the
   * buffer here */

  set_SW_xferlen(bufcom(us->buffer), (window_parameter_data_block.size +
			  window_descriptor_block.size));
  docmd(set_window.size + window_parameter_data_block.size +
	window_descriptor_block.size, us->buffer, 0, NULL, GOOD);
  if (us->verbose) {
    fprintf(stderr, "window(s) set\n");
  }
  return 0;
}

int t630_start_scan(US *us) {
  int widlistlen = 0; /* length of window id list */

  scsicpy(us->buffer, scan.cmd, scan.size);
  set_SC_xfer_length(bufcom(us->buffer), 1);
  set_SC_quality(bufcom(us->buffer), 0);
  docmd(scan.size + widlistlen, us->buffer, 0, NULL, GOOD);
  return 0;
}


int t630_read_data(US *us, int color) {
  unsigned int data_left = 0;
  unsigned int data_read = 0;
  unsigned int data_to_read;
  unsigned int row_bufsize;
  int ret;

  if (us->verbose) {
    fprintf(stderr, "Starting to READ data\n");
  }
  switch(us->colormode) {
   case LINEART:
    data_left = (us->width_in_pixels * us->height_in_pixels / 8);
    break;
   case GREYSCALE:
    /* one byte per pixel */
    data_left = us->width_in_pixels * us->height_in_pixels;
    break;
   case RGB:
    data_left = us->width_in_pixels * us->height_in_pixels;
    break;
  }

  row_bufsize = trim_bufsize(us->bufsize, us->row_len);
  if (us->verbose) {
    fprintf(stderr, "reading %u bytes in rows of %u\n",
	    data_left, row_bufsize);
  }

  do {
    /* if we want to, we can check to see if data_left agrees
     * with amount returned by GET DATA BUFFER STATUS, but why
     * bother? */

    data_to_read = (data_left < row_bufsize) ? data_left : row_bufsize;

    WAIT_SCANNER;
    scsicpy(us->buffer, sread.cmd, sread.size);
    set_R_xfer_length(bufcom(us->buffer), data_to_read);
    ret = do_scsi_cmd(sread.size, us->buffer, data_to_read, us->buffer, GOOD);
    if (ret < 0) {
      if (ret == -4) {
	/* changed to smaller buffer size; try again */
        us->bufsize = -1; /* don't set bufsize back to original size! */
	scsi_init(0, &(us->bufsize));
        row_bufsize = trim_bufsize(us->bufsize, us->row_len);
	if (us->verbose) {
	  fprintf(stderr, "Retrying read: reading %u bytes in rows of %u\n",
		  data_left, row_bufsize);
	}
	continue;
      } else {
	/* some other non-recoverable error */
	return -1;
      }
    }
    t630_write_data(us, color, bufcom(us->buffer), data_to_read, data_read);

    data_left -= data_to_read;
    data_read += data_to_read;
    if (us->verbose) {
      fprintf(stderr, "Buffer of %d bytes read; %d bytes to go\n",
              data_to_read, data_left);
    }
  } while (data_left);
  return 0;
}



#define try_to(command, error) if(command) { fprintf(stderr, error); return -1; }

int vista_s6_do_scan(US *us) {
  try_to(vista_s6_set_window_param(us), "Could not initialize scanner\n")
  try_to(vista_s6_start_scan(us), "Could not start scan\n");
  try_to(vista_s6_read_data(us), "Failed while attempting to read data\n");
  return 0;
}

int vista_s12_do_scan(US *us) {
  try_to(vista_s12_set_window_param(us), "Could not initialize scanner\n")
  try_to(vista_s12_start_scan(us), "Could not start scan\n");
  try_to(vista_s12_read_data(us), "Failed while attempting to read data\n");
  return 0;
}

int t630_do_scan(US *us) {
  int size = us->width_in_pixels * us->height_in_pixels;
  int slop = 1024; /* for header */
  size_t pagesize;
  size_t mmap_size;
  int i;

  if (us->verbose) {
    fprintf(stderr, "size before correction = %d\n", size);
  }
  switch(us->colormode) {
   case LINEART:
    size /= 8;
    break;
   case RGB:
    size *= 3;
    break;
  }
  if (us->verbose) {
    fprintf(stderr, "size after correction = %d\n", size);
  }
  pagesize = getpagesize();
  mmap_size = size+slop + pagesize-((size+slop)%pagesize);
  ftruncate(us->ofd, mmap_size);
  us->arena = mmap(NULL, mmap_size, PROT_READ|PROT_WRITE, MAP_FILE|MAP_SHARED,
                   us->ofd, 0);
  /* If we couldn't mmap, then we need to simply output three pgm's
   * which some external program can combine.  us->arena == 0 tells
   * the write routine to do that.
   * The -1 should be MAP_FAILED, but that doesn't appear in the
   * header files.  Oh, well...
   */
  if (us->arena == (void *) -1) us->arena = 0;
  if (us->arena == (void *) -1) us->arena = NULL;
  if (us->verbose && ! us->arena) {
    fprintf(stderr, "mmap failed, errno=%d: %s\n", errno, strerror(errno));
  }
  switch(us->colormode) {
   case LINEART:
   case GREYSCALE:
    try_to(t630_set_window_param(us, 0), "Could not initialize scanner\n");
    try_to(t630_write_header(us, 0), "Could not write header\n");
    try_to(t630_start_scan(us), "Could not start scan\n");
    try_to(t630_read_data(us, 0), "Failed while attempting to read data\n");
    break;
   case RGB:
    for (i = 1; i <=3; i++) {
      try_to(t630_set_window_param(us, i), "Could not initialize scanner\n");
      try_to(t630_write_header(us, i), "Could not write header\n");
      try_to(t630_start_scan(us), "Could not start scan\n");
      try_to(t630_read_data(us, i), "Failed while attempting to read data\n");
    }
    break;
  }
  if (us->arena) munmap(us->arena, mmap_size);
  /* Now we know the true size, so we can truncate exactly */
  if (us->verbose) {
    fprintf(stderr, "size = %d, offset = %d\n", size, us->arena_offset);
  }
  ftruncate(us->ofd, size + us->arena_offset);
  return 0;
}



/*** Helper function for open_scanner() ***/

/* bound_resolution limits the resolution to what the scanner is capable of */
int bound_resolution(US *us, int x_res, int y_res) {
  if (us->x_resolution > x_res) {
    us->x_resolution = x_res;
  }
  if (us->y_resolution > y_res) {
    us->y_resolution = y_res;
  }
  return 0;
}

/* bound_size limits the size to what the scanner can scan.
 * this is particularly important because the scanners don't have
 * built-in checks and will happily grind their gears if this is exceeded.
 */
int bound_size(US *us, double inches_wide, double inches_tall) {
  if (us->width_in_pixels < 0) {
    us->width_in_pixels = us->x_resolution * inches_wide;
  }
  if (us->height_in_pixels < 0) {
    us->height_in_pixels = us->y_resolution * inches_tall;
  }

  /* bound width and height to stay within scanner, so that
   * -R n for n < resolution doesn't confuse the scanner */
  us->width_in_pixels = min(us->width_in_pixels,
			(us->x_resolution * inches_wide));
  us->height_in_pixels = min(us->height_in_pixels,
			 (us->y_resolution * inches_tall));

  /* Now calculate the absolute width */
  us->phys_width = 1200 * us->width_in_pixels / us->x_resolution;
  us->phys_height = 1200 * us->height_in_pixels / us->y_resolution;
  return 0;
}

/* bound_linelengths_for_colormode gives sane line lengths for each
 * color model.  It might not be strictly necessary; I'm not sure.
 */
int bound_linelengths_for_colormode(US *us) {
  switch(us->colormode) {
   case LINEART:
    /* make sure line width divides evenly by 8... */
    us->width_in_pixels -= us->width_in_pixels % 8;
    us->phys_width = 1200 * us->width_in_pixels / us->x_resolution;
    us->row_len = (us->width_in_pixels / 8);
    if (us->brightness == -1)
      us->brightness = 0x20;
    if (us->threshold == -1)
      us->threshold = 0xe0;
    break;
   case GREYSCALE:
    /* one byte per pixel */
    us->row_len = us->width_in_pixels;
    if (us->brightness == -1)
      us->brightness = 0;
    if (us->threshold == -1)
      us->threshold = 0x80;
    break;
   case RGB:
    /* three bytes per pixel.  Note that high-end scanners
     * might have 6, if they are 30 bits deep total, since
     * they spread each pixel over two bytes.  No weird packing!
     * If you have one of these scanners, please work with me
     * to extend this program to deal with your extra bits.
     */
    us->row_len = us->width_in_pixels * 3;
    if (us->brightness == -1)
      us->brightness = 0;
    if (us->threshold == -1)
      us->threshold = 0x80;
    break;
  }
  return 0;
}


/* open_scanner tries to open the scanner.  It returns 0 on success,
 * and negative numbers on failure.
 * If no device name is specified on the command line, it hunts for
 * the first UMAX scanner on an sg device.
 * If a device name is specified on the command line, it trusts it
 * and does not test it as long as it can be opened.
 * If verbose output is enabled and it is searching for a scanner,
 * it tells the world where it found the scanner.
 */
int open_scanner(US *us) {
  int sfd;
  int i;
  char filename[64]; /* only needs to hold "/dev/sg?"; this is overkill */
  char vendor[9];
  char product[0x10];
  char version[5];
  char s[] = "abcdefgh01234567";
  char *pp;

  /* probe for scanner */
  for (i = 0; i < sizeof(s)-1; i++) {
    if (us->devicename && strlen(us->devicename)) {
      /* user has specified name of device on command line, assume correct */
      sfd = open(us->devicename, O_RDWR);
      if (sfd == -1) { /* unsuccessful*/
        return -1;
      }
    } else {
      sprintf(filename, "/dev/sg%c", s[i]);
      sfd = open(filename, O_RDWR);
      if (sfd == -1) {
        continue;
      }
    }
          
    scsi_init(sfd, &(us->bufsize));
    fullcmd(us->buffer, inquiry, inquiry_return_size, us->buffer, -1);
    get_inquiry_vendor(bufcom(us->buffer), vendor);
    get_inquiry_product(bufcom(us->buffer), product);
    get_inquiry_version(bufcom(us->buffer), version);
    if (strncmp("UMAX    ", vendor, 8)) {
      /* Not a UMAX product, try again. */
      close (sfd);
      continue;
    }
    /* in each case, first put a known space in, then replace
     * the ending spaces with null bytes to end the string. */
  
    pp = &product[0xf];
    product[0xf] = ' ';
    /* leave a space at the end of the version string to use
     * for comparisons when detecting scanners.  We need, for
     * example, to be able to differentiate between "...S6" and
     * "...S6E" */
    while (*(pp - 1) == ' ') *pp-- = '\0';
  
    pp = &version[4];
    version[4] = ' ';
    while (*pp == ' ') *pp-- = '\0';
  
    if (us->verbose) {
      fprintf(stderr, "Found UMAX scanner %sversion %s on device %s\n",
                      product, version, filename);
    }
  
    /* Now identify scanner and fill in the us structure */
    if (!strncmp(product, "Vista-S6 ", 9)) {
      us->grab_scanner = umax_grab_scanner;
      us->do_scan = vista_s6_do_scan;
      us->give_scanner = umax_give_scanner;

      bound_resolution(us, VISTA_S6_X_RESOLUTION, VISTA_S6_Y_RESOLUTION);
      bound_size(us, VISTA_S6_INCHES_WIDE, VISTA_S6_INCHES_TALL);
      bound_linelengths_for_colormode(us);
    } else if (!strncmp(product, "Vista-S8 ", 9)) {
      us->grab_scanner = umax_grab_scanner;
      /* Vista-S8 is a lot like the s6 */
      us->do_scan = vista_s6_do_scan;
      us->give_scanner = umax_give_scanner;

      bound_resolution(us, VISTA_S8_X_RESOLUTION, VISTA_S8_Y_RESOLUTION);
      bound_size(us, VISTA_S8_INCHES_WIDE, VISTA_S8_INCHES_TALL);
      bound_linelengths_for_colormode(us);
    } else if (!strncmp(product, "SuperVista S-12 ", 16)) {
      us->grab_scanner = umax_grab_scanner;
      us->do_scan = vista_s12_do_scan;
      us->give_scanner = umax_give_scanner;

      bound_resolution(us, VISTA_S12_X_RESOLUTION, VISTA_S12_Y_RESOLUTION);
      bound_size(us, VISTA_S12_INCHES_WIDE, VISTA_S12_INCHES_TALL);
      bound_linelengths_for_colormode(us);
    } else if (!strncmp(product, "UC630 ", 5)) {
      /* NOTE: There may be other product ID's.  Please tell me so
       * I can fix this driver...
       */
      us->grab_scanner = umax_grab_scanner;
      us->do_scan = t630_do_scan;
      us->give_scanner = umax_give_scanner;

      bound_resolution(us, T630_X_RESOLUTION, T630_Y_RESOLUTION);
      bound_size(us, T630_INCHES_WIDE, T630_INCHES_TALL);
      if (us->threshold == -1)
        us->threshold = 0x20;
      bound_linelengths_for_colormode(us);
    } else if (!strncmp(product, "UG630 ", 5)) {
      /* greyscale-only scanner
       */
      if (us->colormode == RGB) {
        fprintf(stderr, "%sis a greyscale scanner.\n", product);
        return -3;
      }

      us->grab_scanner = umax_grab_scanner;
      us->do_scan = t630_do_scan;
      us->give_scanner = umax_give_scanner;

      bound_resolution(us, T630_X_RESOLUTION, T630_Y_RESOLUTION);
      bound_size(us, T630_INCHES_WIDE, T630_INCHES_TALL);
      if (us->threshold == -1)
        us->threshold = 0x20;
      bound_linelengths_for_colormode(us);
    } else {
      /* we don't recognize this scanner yet! */
      fprintf(stderr, "UMAX scanner %s version %s on device %s "
                      "is currently an unrecognized device.\n"
                      "Please add explicit support for it to umax.c\n",
                      product, version, filename);
      continue;
    }
    if (us->verbose) {
      fprintf(stderr,
		  "Size- and position-related variables:\n"
		  "row_len = %u (measured in bytes)\n"
		  "width_in_pixels = %u\n"
		  "height_in_pixels = %u\n"
		  "phys_width = %u (1200pt/inch)\n"
		  "phys_height = %u (1200pt/inch)\n"
		  "x_resolution = %u (pixels/inch)\n"
		  "y_resolution = %u (pixels/inch)\n"
		  "upper_left_x = %u (offset 1200pt/inch)\n"
		  "upper_left_y = %u (offset 1200pt/inch)\n"
		  "brightness = %u, threshold = %u\n"
		  "\n",
		  us->row_len, us->width_in_pixels, us->height_in_pixels,
		  us->phys_width, us->phys_height,
		  us->x_resolution, us->y_resolution,
		  us->upper_left_x, us->upper_left_y,
		  us->brightness, us->threshold
		  );
    }
    /* At this point, if we haven't found the scanner, we have
     * "continue"d to the beginning of the loop, and if we've
     * exhausted the potential sg devices, we've terminated the
     * loop.  So at this point, we have found the scanner and
     * done everything we want to with it... */
    return 0;
  }

  /* if we reach here, we haven't found a UMAX scanner on any device */
  if (us->verbose) {
    fprintf(stderr, "No UMAX scanner found\n");
  }
  return -2;
}



void usage(char * errormessage) {
  fprintf(stderr, "%s\n", errormessage);
  fprintf(stderr, "Usage: umax -lgr|-VH [options] [devicename]\n");
  fprintf(stderr, "l = lineart, g = greyscale, r = rgb\n"
		  "V = Version, H = Help\n"
		  "Options:\n"
	          "v = verbose, vv = very verbose (scsi command debugging)\n"
	          "p = preview quality (not yet implemented)\n"
	          "a = use automatic document feeder (adf)\n"
	          "x = x resolution, y = y resolution, R = Resolution (both)\n"
	          "w = width in pixels, h = height in pixels\n"
	          "X = ULX, Y = ULY (1200 units per inch)\n"
	          "b = brightness (0), t = threshold (0x80)\n"
	          "[devicename] is optional; /dev/sg[a-f0-7] are searched in order.\n"
	  );
  exit(1);
}


/* Should call functions in this order:
 * open_scanner
 * us.grab_scanner
 * us.do_scan
 * us.give_scanner
 *
 * us.do_scan should call:
 *   set_window_param
 *   start_scan
 *   read_data
 * (or equivalent depending on the scanner)
 */

int main (int argc, char **argv) {
  int debug_out;
  char c;
  US us;

  /* Initialize us structure */
  us.bufsize = -1;
  us.devicename = NULL;
  us.verbose = 0;
  us.colormode = 0; /* LINEART, GREYSCALE, or RGB */
  us.row_len = -1;
  us.width_in_pixels = -1;
  us.height_in_pixels = -1;
  us.phys_width = -1; /* 1200pt/inch */
  us.phys_height = -1; /* 1200pt/inch */
  us.max_value = 255;
  us.x_resolution = 300;
  us.y_resolution = 300;
  us.upper_left_x = 0; /* 1200pt/inch */
  us.upper_left_y = 0; /* 1200pt/inch */
  us.brightness = -1;
  us.threshold = -1;
  us.ofd = STDOUT_FILENO; /* output file descriptor, stdout by default */

  if(argc < 2) {
    usage("too few arguments");
  }
  while ((c = getopt(argc, argv, "VHlgrvpax:y:R:w:h:X:Y:b:t:o:")) != EOF) {
    switch (c) {
    case 'V':
      usage("Umax driver version 0.5\n");
    case 'H':
      usage("Umax help:");
    case 'l':
      us.colormode = LINEART;
      break;
    case 'g':
      us.colormode = GREYSCALE;
      break;
    case 'r':
      us.colormode = RGB;
      break;
    case 'v':
      us.verbose++;
      break;
    case 'p':
      /* future preview mode */
      break;
    case 'a':
      us.adf = 1;
      break;
    case 'x':
      us.x_resolution = atoi(optarg);
      break;
    case 'y':
      us.y_resolution = atoi(optarg);
      break;
    case 'R':
      us.x_resolution = us.y_resolution = atoi(optarg);
      break;
    case 'w':
      us.width_in_pixels = atoi(optarg);
      break;
    case 'h':
      us.height_in_pixels = atoi(optarg);
      break;
    case 'X':
      us.upper_left_x = atoi(optarg);
      break;
    case 'Y':
      us.upper_left_y = atoi(optarg);
      break;
    case 'b':
      us.brightness = atoi(optarg);
      break;
    case 't':
      us.threshold = atoi(optarg);
      break;
    case 'o':
      us.ofd = open(optarg, O_CREAT|O_TRUNC|O_RDWR,
                    S_IRUSR|S_IWUSR | S_IRGRP|S_IWGRP | S_IROTH|S_IWOTH);
      break;
    case 0x3f:
      fprintf(stderr, "getopt complains about bad use of option :-(\n"
	              "Trying to continue anyway...\n");
      break;
    default:
      fprintf(stderr, "Unknown option '-%c' (ascii %02xh)\n", c, c);
      usage("bad option");
    }
  }
  if(optind < argc) {
    us.devicename = argv[optind];
  }

  if (us.colormode == 0) {
    usage ("Must choose lineart, greyscale, or rgb");
  }

  if (us.verbose > 1) {
    debug_out = open("/tmp/umax.scsi.debug", O_WRONLY|O_CREAT|O_APPEND,
                     S_IRUSR|S_IWUSR | S_IRGRP|S_IWGRP | S_IROTH|S_IWOTH);
    if (!debug_out) {
      /* open failed */
      usage ("Couldn't open /tmp/umax.scsi.debug");
    }
    scsi_init_debug(debug_out);
  }


#define try(command, error) if(command) { fprintf(stderr, error); exit(1); }
  try (open_scanner(&us), "Could not open scanner\n");
  try (us.grab_scanner(&us), "Could not allocate scanner\n");
  try (us.do_scan(&us), "Could not do scan\n");
  if (us.give_scanner(&us)) {
    /* does not need to abort; got the scan, but bus or scanner is
     * probably hung... */
    fprintf (stderr, "Failed to deallocate scanner\n");
  }

  exit (0);
}
